from base_st_classes import *

class LookupSuite(NFSSuite):
    """Test operation 15: LOOKUP

    Equivalence partitioning:

    Input Condition: current filehandle
        Valid equivalence classes:
            dir(10)
        Invalid equivalence classes:
            not directory or symlink(11)
            invalid filehandle(12)
            symlink(13)
    Input Condition: objname
        Valid equivalence classes:
            legal name(20)
        Invalid equivalence classes:
            zero length(21)
            non-utf8(22)
            non-existent object(23)
            non-accessible object(24)
    """
    def tearDown(self):
        try: self.clean_dir(self.tmp_dir)
        except SkipException, e:
            pass
    #
    # Testcases covering valid equivalence classes.
    #
    def testDir(self):
        """LOOKUP simple test

        Covered valid equivalence classes: 10, 20
        """
        ops = [self.ncl.putrootfh_op()] + self.ncl.lookup_path(self.dirfile)
        res = self.ncl.do_ops(ops)
        self.assert_OK(res)

    #
    # Testcases covering invalid equivalence classes.
    #
    def testFhNotDir(self):
        """LOOKUP with non-dir (cfh) should give NFS4ERR_NOTDIR

        Covered invalid equivalence classes: 11
        """
        for path in [ self.regfile,
                      self.blockfile,
                      self.charfile,
                      self.socketfile,
                      self.fifofile ]:
            lookupops = self.ncl.lookup_path(path)
            ops = [self.ncl.putrootfh_op()] + lookupops
            ops.append(self.ncl.lookup_op("porting"))
            res = self.ncl.do_ops(ops)
            self.assert_status(res, [NFS4ERR_NOTDIR])

    def testSymlinkFh(self):
        """LOOKUP with (cfh) as symlink should return NFS4ERR_SYMLINK

        Covered invalid equivalence classes: 13
        """
        lookupops = self.ncl.lookup_path(self.dirsymlinkfile)
        ops = [self.ncl.putrootfh_op()] + lookupops
        ops.append(self.ncl.lookup_op("porting"))
        res = self.ncl.do_ops(ops)
        self.assert_status(res, [NFS4ERR_SYMLINK])

    def testNoFh(self):
        """LOOKUP without (cfh) should return NFS4ERR_NOFILEHANDLE

        Covered invalid equivalence classes: 12
        """
        lookupops = self.ncl.lookup_path(self.regfile)
        res = self.ncl.do_ops(lookupops)
        self.assert_status(res, [NFS4ERR_NOFILEHANDLE])

    def testNonExistent(self):
        """LOOKUP with non-existent components should return NFS4ERR_NOENT

        Covered invalid equivalence classes: 23
        """
        lookupops = self.ncl.lookup_path(self.vaporfile)
        ops = [self.ncl.putrootfh_op()] + lookupops
        res = self.ncl.do_ops(ops)
        self.assert_status(res, [NFS4ERR_NOENT])

    def testNonAccessable(self):
        """LOOKUP with non-accessable components should return NFS4ERR_ACCESS

        Covered invalid equivalence classes: 24
        """
        lookupops = self.ncl.lookup_path(self.notaccessablefile)
        ops = [self.ncl.putrootfh_op()] + lookupops
        ops.append(self.ncl.getfh_op())
        res = self.ncl.do_ops(ops)
        self.assert_status(res, [NFS4ERR_ACCESS])

    def testNonUTF8(self):
        """LOOKUP with non-UTF8 name should return NFS4ERR_INVAL

        Covered invalid equivalence classes: 22
        """
        for name in self.get_invalid_utf8strings():
            ops = [self.ncl.putrootfh_op()]
            ops.append(self.ncl.lookup_op(name))
            res = self.ncl.do_ops(ops)
            self.assert_status(res, [NFS4ERR_INVAL])

    def testZeroLength(self):
        """LOOKUP with zero length name should return NFS4ERR_INVAL

        Covered invalid equivalence classes: 21
        """
        ops = [self.ncl.putrootfh_op()]
        ops.append(self.ncl.lookup_op(""))
        res = self.ncl.do_ops(ops)
        self.assert_status(res, [NFS4ERR_INVAL])

    def testLongName(self):
        """LOOKUP should fail with NFS4ERR_NAMETOOLONG with long filenames"""
        longname = "a"*512
        ops = [self.ncl.putrootfh_op()]
        ops.append(self.ncl.lookup_op(longname))
        res = self.ncl.do_ops(ops)
        self.assert_status(res, [NFS4ERR_NAMETOOLONG])
        
    #
    # Extra tests.
    #
    def _assert_noent(self, pathcomps):
        lookupops = self.ncl.lookup_path(pathcomps)
        ops = [self.ncl.putrootfh_op()] + lookupops
        res = self.ncl.do_ops(ops)
        self.assert_status(res, [NFS4ERR_NOENT])

    def testDots(self):
        """LOOKUP on (nonexistent) . and .. should return NFS4ERR_NOENT 

        Extra test

        Comments: Even if the server does not allow creation of files
        called . and .., LOOKUP should return NFS4ERR_NOENT. 
        """
        # Try lookup on ["doc", ".", "README"]
        # First, make sure there is no object named "." in the doc directory
        # Note: The file doc/./README actually exists on a UNIX server
        self._assert_noent(self.dirfile + ["."])
        self._assert_noent(self.dirfile + [".", "README"])
        
        # Same goes for ".."
        # Note: The file doc/porting/../README actually exists on a UNIX server
        self._assert_noent(self.docporting + [".."])
        self._assert_noent(self.docporting + ["..", "README"])

    def testValidNames(self):
        """LOOKUP should succeed on all legal names

        Extra test

        Comments: This test tries LOOKUP on all names returned from try_file_names()
        """
        self.init_connection()

        # Saved files for LOOKUP
        try:
            (accepted_names, rejected_names) = self.try_file_names(0)
        except SkipException, e:
            self.skip(e)

        # Ok, lets try LOOKUP on all accepted names
        lookup_dir_ops = self.ncl.lookup_path(self.tmp_dir)
        for filename in accepted_names:
            ops = [self.ncl.putrootfh_op()] + lookup_dir_ops
            ops.append(self.ncl.lookup_op(filename))
            res = self.ncl.do_ops(ops)
            self.assert_OK(res)
            
    def testInvalidNames(self):
        """LOOKUP should fail with NFS4ERR_NOENT on all unexisting, invalid file names

        Extra test

        (FRED) - Below is Peter's comment, but I disagree, and have changed
        test accordingly.
        
        Comments: Tries LOOKUP on rejected file names from
        try_file_names().  NFS4ERR_INVAL should NOT be returned in this case,
        although the server rejects creation of objects with these names
        """
        self.init_connection()

        try:
            (accepted_names, rejected_names) = self.try_file_names()
        except SkipException, e:
            self.skip(e)

        # Ok, lets try LOOKUP on all rejected names
        lookup_dir_ops = self.ncl.lookup_path(self.tmp_dir)
        for filename in rejected_names:
            ops = [self.ncl.putrootfh_op()] + lookup_dir_ops
            ops.append(self.ncl.lookup_op(filename))
            res = self.ncl.do_ops(ops)
            self.assert_status(res, [NFS4ERR_INVAL,NFS4ERR_NOENT])
