from base_st_classes import *

class OpendowngradeSuite(NFSSuite):
    """Test operation 21: OPEN_DOWNGRADE

    Equivalence partitioning:
    
    Input Condition: current filehandle
        Valid equivalence classes:
            file(1)
            named attribute(2)
        Invalid equivalence classes:
            dir(3)
            special device files(4)
            invalid filehandle(10)
    Input Condition: stateid
        Valid equivalence classes:
            all bits zero(17)
            all bits one(18)
            valid stateid from open(19)
        Invalid equivalence classes:
            invalid stateid(20)
    Input Condition: seqid
        Valid equivalence classes:
            correct seqid(10)
        Invalid equivalence classes:
            too small seqid(11)
            too large seqid(12)
    Input Condition: share_access
        Valid equivalence classes:
            valid share_access(30)
        Invalid equivalence classes:
            invalid share_access(31)
    Input Condition: share_deny
        Valid equivalence classes:
            valid share_deny(40)
        Invalid equivalence classes:
            invalid share_deny(41)

    """

    def setUp(self):
        NFSSuite.setUp(self)
        self.fh1 = self.fh2 = None
        self.state1 = self.state2 = None
        self.dir = self.regfile[:-1]
        self.name = self.regfile[-1]
        self.init_connection()

    def tearDown(self):
        self.ncl.close_file(self.fh1, self.state1, error=None)
        self.ncl.close_file(self.fh2, self.state2, error=None)
        try: self.clean_dir(self.tmp_dir)
        except SkipException, e:
            pass

    def _doOpenDowngrade(self, fh, stateid, seqid=None,
                         access=OPEN4_SHARE_ACCESS_READ,
                         deny=OPEN4_SHARE_DENY_NONE, error=[NFS4_OK]):
        if seqid==None:
            seqid = self.ncl.get_seqid()
        putfhop = self.ncl.putfh_op(fh)
        opendgop = self.ncl.open_downgrade_op(open_stateid=stateid, seqid=seqid, share_access=access, share_deny=deny)
        res = self.ncl.do_ops([putfhop, opendgop])
        self.assert_status(res,error)
        if res.status == NFS4_OK:
            # FRED - note if NFS4_OK is not in error, then will never return
            #        new stateid
            return res.resarray[-1].arm.arm.open_stateid
        
    def __closefile(self,fh,stateid,error=[NFS4_OK]):
        """close the given file
        """
        ops = [self.ncl.putfh_op(fh)]
        ops.append(self.ncl.close_op(self.ncl.get_seqid(),stateid))
        res = self.ncl.do_ops(ops)
        self.assert_status(res,error)

    #
    # Testcases covering valid equivalence classes.
    #
    def testRegularOpen(self):
        """OPENDOWNGRADE on regular file
    
        Covered valid equivalence classes: 1, 10, 19, 30, 40
    	"""
        self.fh1, self.state1 = \
                  self.ncl.open_confirm(self.dir, self.name,
                                        access=OPEN4_SHARE_ACCESS_READ,
                                        owner="RegularOpen")
        self.fh1, self.state1 = \
                  self.ncl.open_confirm(self.dir, self.name,
                                        access=OPEN4_SHARE_ACCESS_BOTH,
                                        owner="RegularOpen")
        self.state1 = self._doOpenDowngrade(self.fh1, self.state1)

    def testNewState1(self):
        """OPENDOWNGRADE to new state should return NFS4ERR_INVAL
    	"""

        self.fh1, self.state1 = \
                  self.ncl.open_confirm(self.dir, self.name,
                                        access=OPEN4_SHARE_ACCESS_BOTH)
        self._doOpenDowngrade(self.fh1, self.state1, error=[NFS4ERR_INVAL])

    def testNewState2(self):
        """OPENDOWNGRADE to new state should return NFS4ERR_INVAL
    	"""

        self.fh1, self.state1 = \
                  self.ncl.open_confirm(self.dir, self.name,
                                        access=OPEN4_SHARE_ACCESS_WRITE,
                                        owner="NewState2")
        self.fh1, self.state1 = \
                  self.ncl.open_confirm(self.dir, self.name,
                                        access=OPEN4_SHARE_ACCESS_BOTH,
                                        owner="NewState2")
        self._doOpenDowngrade(self.fh1, self.state1, error=[NFS4ERR_INVAL])

    def testBadSeqid(self):
        """OPENDOWNGRADE with bad seqid should return NFS4ERR_BAD_SEQID
    	"""

        self.fh1, self.state1 = \
                  self.ncl.open_confirm(self.dir, self.name)

        seqid = (self.ncl.seqid+2) % 2**32L
        self._doOpenDowngrade(self.fh1, self.state1, seqid,
                              error=[NFS4ERR_BAD_SEQID])

    def testBadStateid(self):
        """OPENDOWNGRADE with bad stateid should return NFS4ERR_BAD_STATEID"""
        
        self.fh1, self.state1 = \
                  self.ncl.open_confirm(self.dir, self.name,
                                        access=OPEN4_SHARE_ACCESS_BOTH)
        stateid = stateid4(0, "")
        self._doOpenDowngrade(self.fh1, stateid, (self.ncl.seqid+1) % 2**32L,
                              error=[NFS4ERR_BAD_STATEID])

    def testOldStateid(self):
        """OPENDOWNGRADE with old stateid should return NFS4ERR_OLD_STATEID"""

        # OPEN
        ops = [self.ncl.putrootfh_op()] + self.ncl.lookup_path(self.dir)
        ops.append(self.ncl.open(self.name))
        ops.append(self.ncl.getfh_op())
        res = self.ncl.do_ops(ops)
        self.assert_OK(res)
        fh = res.resarray[-1].arm.arm.object
        oldstateid = res.resarray[-2].arm.arm.stateid

        # CONFIRM
        ops = [self.ncl.putfh_op(fh)]
        confirmop = self.ncl.open_confirm_op(oldstateid,self.ncl.get_seqid())
        ops.append(confirmop)
        res = self.ncl.do_ops(ops)
        self.assert_OK(res)
        self.fh1 = fh
        self.state1 = res.resarray[-1].arm.arm.open_stateid

        self._doOpenDowngrade(fh, oldstateid, error=[NFS4ERR_OLD_STATEID])

    def testNoCfh(self):
        """OPENDOWNGRADE with no (cfh) should return NFS4ERR_NOFILEHANDLE"""

        self.fh1, self.state1 = \
                  self.ncl.open_confirm(self.dir, self.name,
                                        access=OPEN4_SHARE_ACCESS_BOTH)

        opendgop=self.ncl.open_downgrade_op(open_stateid=self.state1,
                                            seqid=self.ncl.get_seqid(),
                                            share_access=OPEN4_SHARE_ACCESS_READ,
                                            share_deny=OPEN4_SHARE_DENY_NONE)
        res = self.ncl.do_ops([opendgop])
        self.assert_status(res,[NFS4ERR_NOFILEHANDLE])
 
