from base_st_classes import *

class SecinfoSuite(NFSSuite):
    """Test operation 33: SECINFO

    Equivalence partitioning:

    Input Condition: current filehandle
        Valid equivalence classes:
            dir(10)
        Invalid equivalence classes:
            not dir(11)
            invalid filehandle(12)
    Input Condition: name
        Valid equivalence classes:
            valid name(20)
        Invalid equivalence classes:
            non-existent object(21)
            zerolength(22)
            non-utf8(23)

    Comments: It's hard to cover eq. class 12, since it's not possible
    PUTFH an invalid filehandle. 
    """
    #
    # Testcases covering valid equivalence classes.
    #
    def testValid(self):
        """SECINFO on existing file

        Covered valid equivalence classes: 10, 20
        """
        ops = [self.ncl.putrootfh_op()] + self.ncl.lookup_path(self.dirfile)
        ops.append(self.ncl.secinfo_op("README"))

        res = self.ncl.do_ops(ops)
        self.assert_OK(res)

        # Make sure at least one security mechanisms is returned.
        mechanisms = res.resarray[-1].arm.arm
        self.failIf(len(mechanisms) < 1,
                    "SECINFO returned no security mechanisms")

    #
    # Testcases covering invalid equivalence classes.
    #
    def testFhNotDir(self):
        """SECINFO with non-dir (cfh) should give NFS4ERR_NOTDIR

        Covered invalid equivalence classes: 11
        """
        ops = [self.ncl.putrootfh_op()] + self.ncl.lookup_path(self.regfile)
        ops.append(self.ncl.secinfo_op("README"))

        res = self.ncl.do_ops(ops)
        self.assert_status(res, [NFS4ERR_NOTDIR])

    def testNonExistent(self):
        """SECINFO on non-existing object should return NFS4ERR_NOENT

        Covered invalid equivalence classes: 21
        """
        ops = [self.ncl.putrootfh_op()] + self.ncl.lookup_path(self.dirfile)
        ops.append(self.ncl.secinfo_op(self.vaporfilename))

        res = self.ncl.do_ops(ops)
        self.assert_status(res, [NFS4ERR_NOENT])

    def testZeroLengthName(self):
        """SECINFO with zero length name should return NFS4ERR_INVAL

        Covered invalid equivalence classes: 22
        """
        ops = [self.ncl.putrootfh_op()] + self.ncl.lookup_path(self.dirfile)
        ops.append(self.ncl.secinfo_op(""))

        res = self.ncl.do_ops(ops)
        self.assert_status(res, [NFS4ERR_INVAL])

    def testNonUTF8(self):
        """SECINFO with non-UTF8 name should return NFS4ERR_INVAL

        Covered invalid equivalence classes: 23

        Comments: It does'nt matter that the files does not exist; the UTF8
        check should return NFS4ERR_INVAL anyway. 
        """
        for name in self.get_invalid_utf8strings():
            ops = [self.ncl.putrootfh_op()] 
            ops.append(self.ncl.secinfo_op(name))

            res = self.ncl.do_ops(ops)
            self.assert_status(res, [NFS4ERR_INVAL])

    #
    # Extra tests. 
    #
    def testRPCSEC_GSS(self):
        """SECINFO must return at least RPCSEC_GSS

        Extra test
        """
        # FIXME: Also verify that all Kerberos and LIPKEY security triples
        # listed in section 3.2.1.1 and 3.2.1.2 are supported.
        lookupops = self.ncl.lookup_path(self.dirfile)
        ops = [self.ncl.putrootfh_op()] + lookupops
        ops.append(self.ncl.secinfo_op("README"))

        res = self.ncl.do_ops(ops)
        self.assert_OK(res)
        mechanisms = res.resarray[-1].arm.arm
        found_rpcsec_gss = 0

        for mech in mechanisms:
            if mech.flavor == rpc.RPCSEC_GSS:
                found_rpcsec_gss = 1

        self.failIf(not found_rpcsec_gss,
                    "SECINFO did not return (mandatory) flavor RPCSEC_GSS")


    def _assert_secinfo_response(self, name):
        lookupops = self.ncl.lookup_path(self.dirfile)
        ops = [self.ncl.putrootfh_op()] + lookupops
        ops.append(self.ncl.secinfo_op(name))
        
        res = self.ncl.do_ops(ops)
        self.assert_status(res, [NFS4ERR_NOENT, NFS4ERR_INVAL])

    def testDots(self):
        """SECINFO on . and .. should return NOENT/INVAL in /doc

        Extra test
        """
        # . should not exist in doc dir
        if not self.make_sure_nonexistent(".", self.dirfile): return
        self._assert_secinfo_response(".")

        # .. should not exist in doc dir
        if not self.make_sure_nonexistent("..", self.dirfile): return
        self._assert_secinfo_response("..")

    def testValidNames(self):
        """SECINFO should succeed on all legal names

        Extra test

        Comments: This test tries SECINFO on all names returned from try_file_names()
        """
        # This test tests the lookup part of SECINFO
        self.init_connection()

        # Save files for SECINFO
        try:
            (accepted_names, rejected_names) = self.try_file_names(0)
        except SkipException, e:
            self.skip(e)

        # Ok, lets try SECINFO on all accepted names
        lookup_dir_ops = self.ncl.lookup_path(self.tmp_dir)
        for filename in accepted_names:
            ops = [self.ncl.putrootfh_op()] + lookup_dir_ops
            ops.append(self.ncl.secinfo_op(filename))
            res = self.ncl.do_ops(ops)
            self.assert_OK(res)
    
    def testInvalidNames(self):
        """SECINFO should fail with NFS4ERR_NOENT on all unexisting, invalid file names
         
        Extra test

        Comments: Tries SECINFO on rejected file names from
        try_file_names().  NFS4ERR_INVAL should NOT be returned in this case,
        although the server rejects creation of objects with these names

        (FRED) - see same test in st_lookup
        """
        self.init_connection()
        try:
            (accepted_names, rejected_names) = self.try_file_names()
        except SkipException, e:
            self.skip(e)

        # Ok, lets try SECINFO on all rejected names
        lookup_dir_ops = self.ncl.lookup_path(self.tmp_dir)
        for filename in rejected_names:
            ops = [self.ncl.putrootfh_op()] + lookup_dir_ops
            ops.append(self.ncl.secinfo_op(filename))
            res = self.ncl.do_ops(ops)
            self.assert_status(res, [NFS4ERR_NOENT,NFS4ERR_INVAL])


# [NOTE] The testValidNames and testInvalidNames tests may be redundant.
#   It is likely that the function on the server that checks for NonUTF8 names also 
#       checks for Valid and Invalid names.
#       *That same function is used by lookup for the same check on the server. 

