from base_st_classes import *

class VerifySuite(NFSSuite):
    """Test operation 37: VERIFY

    Equivalence partitioning:

    Input Condition: current filehandle
        Valid equivalence classes:
            link(1)
            block(2)
            char(3)
            socket(4)
            FIFO(5)
            dir(6)
            file(7)
        Invalid equivalence classes:
            invalid filehandle(8)
    Input Condition: fattr.attrmask
        Valid equivalence classes:
            valid attribute(9)
        Invalid equivalence classes:
            invalid attrmask(10) (FATTR4_*_SET)
    Input Condition: fattr.attr_vals
        Valid equivalence classes:
            same attributes(11)
            not same attributes(12)
        Invalid equivalence classes:
            attr with invalid utf8(13)

    """

    def _verify(self,lookupops,attrdict={FATTR4_SIZE:17},
                error=[NFS4_OK]):
        """call verify and then getattr
        """
        ops = [self.ncl.putrootfh_op()] + lookupops
        attrs = nfs4lib.dict2fattr(attrdict)
        ops.append(self.ncl.verify_op(attrs))
        ops.append(self.ncl.getattr(attrdict.keys()))
        res = self.ncl.do_ops(ops)
        self.assert_status(res,error)
        return res

    #
    # Testcases covering valid equivalence classes.
    #
    def testSame(self):
        """VERIFY with same attributes should execute remaining ops

        Covered valid equivalence classes: 1, 2, 3, 4, 5, 6, 7, 9, 11
        """
        # Fetch sizes for all objects
        obj_sizes = self.ncl.lookup_all_objects_and_sizes()
        
        # For each type of object, do verify with same filesize
        # get filesize again, and check if it matches previous size.
        for (lookupops, objsize) in obj_sizes:
            res = self._verify(lookupops,{FATTR4_SIZE:objsize})

            # File sizes should match. 
            obj = res.resarray[-1].arm.arm.obj_attributes
            d =  nfs4lib.fattr2dict(obj)
            new_size = d[FATTR4_SIZE]
            self.failIf(objsize != new_size,
                        "GETATTR after VERIFY returned different filesize")

    def testNotSame(self):
        """VERIFY with not same attributes should return NFS4ERR_NOT_SAME

        Covered valid equivalence classes: 1, 2, 3, 4, 5, 6, 7, 9, 12
        """
        # Fetch sizes for all objects
        obj_sizes = self.ncl.lookup_all_objects_and_sizes()
        
        # For each type of object, do verify with wrong filesize. 
        for (lookupops, objsize) in obj_sizes:
            self._verify(lookupops, {FATTR4_SIZE:objsize+17},
                         error=[NFS4ERR_NOT_SAME])

    def testType(self):
        """VERIFY file type"""
        lookupops = self.ncl.lookup_path(self.dirfile)
        self._verify(lookupops, {FATTR4_TYPE:NF4DIR})
        
    #
    # Testcases covering invalid equivalence classes.
    #
    def testNoFh(self):
        """VERIFY without (cfh) should return NFS4ERR_NOFILEHANDLE

        Covered invalid equivalence classes: 8
        """
        attrs = nfs4lib.dict2fattr({FATTR4_SIZE:17})
        verifyop = self.ncl.verify_op(attrs)
        res = self.ncl.do_ops([verifyop])
        self.assert_status(res, [NFS4ERR_NOFILEHANDLE])

    def testWriteOnlyAttributes(self):
        """VERIFY with FATTR4_*_SET should return NFS4ERR_INVAL

        Covered invalid equivalence classes: 10

        Comments: See GetattrSuite.testWriteOnlyAttributes. 
        """
        
        lookupops = self.ncl.lookup_path(self.regfile)
        settime = settime4(set_it=SET_TO_SERVER_TIME4)
        for attr in [FATTR4_TIME_ACCESS_SET, FATTR4_TIME_MODIFY_SET]:
            self._verify(lookupops, {attr:settime},
                         error=[NFS4ERR_INVAL])

    def testNonUTF8(self):
        """VERIFY with non-UTF8 FATTR4_OWNER should return NFS4ERR_INVAL

        Covered invalid equivalence classes: 13
        """
        # FRED - NFS4ERR_NOT_SAME seems reasonable
        lookupops = self.ncl.lookup_path(self.regfile)
        for name in self.get_invalid_utf8strings():
            self._verify(lookupops,{FATTR4_OWNER:name},
                         error=[NFS4ERR_INVAL, NFS4ERR_NOT_SAME])

    def testRdattr_error(self):
        """VERIFY with FATTR4_RDATTR_ERROR set should return NFS4ERR_INVAL

        Covered invalid equivalence classes: 10

        Comments: See GetattrSuite.testWriteOnlyAttributes. 
        """
        lookupops = self.ncl.lookup_path(self.regfile)
        self._verify(lookupops,{FATTR4_RDATTR_ERROR:NFS4_OK},
                     error=[NFS4ERR_INVAL])

    def testUnsupported(self):
        """VERIFY with an unsupported attribute returns NFS4ERR_ATTRNOTSUPP
        """

        lookupops = self.ncl.lookup_path(self.regfile)
        for (bitnum,value,name) in self.unsupportedAttrs(self.regfile):
            self._verify(lookupops,{bitnum:value},error=[NFS4ERR_ATTRNOTSUPP])
